import pandas
from ccxt.base.exchange import Exchange
from ccxt.base.errors import BadRequest, BadSymbol
from btrccts.check_dataframe import _check_dataframe
from btrccts.exchange_account import ExchangeAccount


class ExchangeBackend:

    def __init__(self, timeframe, balances={}, ohlcvs={}):
        self._account = ExchangeAccount(timeframe=timeframe,
                                        balances=balances,
                                        ohlcvs=ohlcvs)
        self._ohlcvs = {}
        self._timeframe = timeframe
        for key in ohlcvs:
            self._ohlcvs[key] = _check_dataframe(
                ohlcvs[key],
                timeframe,
                ['open', 'low', 'high', 'close', 'volume'])

    def fetch_order(self, id, symbol=None):
        return self._account.fetch_order(id=id, symbol=symbol)

    def fetch_balance(self):
        return self._account.fetch_balance()

    def create_order(self, market, type, price, side, amount):
        return self._account.create_order(market=market, type=type, side=side,
                                          price=price, amount=amount)

    def cancel_order(self, id, symbol=None):
        return self._account.cancel_order(id=id, symbol=symbol)

    def fetch_open_orders(self, symbol=None, since=None, limit=None):
        return self._account.fetch_open_orders(symbol=symbol, since=since,
                                               limit=limit)

    def fetch_closed_orders(self, symbol=None, since=None, limit=None):
        return self._account.fetch_closed_orders(symbol=symbol, since=since,
                                                 limit=limit)

    def fetch_ticker(self, symbol):
        ohlcv = self._ohlcvs.get(symbol)
        if ohlcv is None:
            raise BadSymbol('ExchangeBackend: no prices for {}'.format(symbol))
        current_date = self._timeframe.date().floor('1T')
        row = ohlcv.loc[current_date]
        timestamp = int(current_date.value / 10**6)
        return {
            'symbol': symbol,
            'timestamp': timestamp,
            'datetime': Exchange.iso8601(timestamp),
            'high': row['high'],
            'low': row['low'],
            'bid': None,
            'bidVolume': None,
            'ask': None,
            'askVolume': None,
            'vwap': None,
            'open': row['open'],
            'close': row['close'],
            'last': None,
            'previousClose': None,
            'change': None,
            'percentage': None,
            'average': None,
            'baseVolume': None,
            'quoteVolume': None,
            'info': {},
        }

    def fetch_ohlcv_dataframe(self, symbol, timeframe='1m', since=None,
                              limit=None, params={}):
        # Exchanges in the real world have different behaviour, when there is
        # no since parameter provided. (some use data from the beginning,
        # some from the end)
        # We return data from the beginning, because this is most likely not
        # what the user wants, so this will force the user to provide the
        # parameters, which will work with every exchange. This is a bug
        # prevention mechanism.
        ohlcv = self._ohlcvs.get(symbol)
        if ohlcv is None:
            raise BadSymbol('ExchangeBackend: no prices for {}'.format(symbol))
        pd_current_date = self._timeframe.date().floor('1T')
        if limit is None:
            limit = 5
        timeframe_sec = Exchange.parse_timeframe(timeframe)
        pd_timeframe = pandas.Timedelta(timeframe_sec, unit='s')
        ohlcv_start_date = ohlcv.index[0]
        if since is None:
            pd_since = ohlcv_start_date
        else:
            pd_since = pandas.Timestamp(since, unit='ms', tz='UTC')
        pd_since = pd_since.ceil(pd_timeframe)
        if pd_since < ohlcv_start_date:
            raise BadRequest('ExchangeBackend: fetch_ohlcv: no date availabe '
                             'at since')
        pd_until = pd_since + limit * pd_timeframe - pandas.Timedelta('1m')
        if pd_until >= pd_current_date + pd_timeframe:
            raise BadRequest(
                'ExchangeBackend: fetch_ohlcv:'
                ' since.ceil(timeframe) + limit * timeframe'
                ' needs to be in the past')
        pd_until = min(pd_until, pd_current_date)
        data = ohlcv[pd_since:pd_until]
        return data.resample(pd_timeframe).agg({
            'open': 'first',
            'high': 'max',
            'low': 'min',
            'close': 'last',
            'volume': 'sum'})



class AlgorithmBaseSync:

    @staticmethod
    def configure_argparser(argparser):
        pass

    def __init__(self, context, args):
        pass

    def exit(self, reason):
        pass

    def handle_exception(self, e):
        raise e

    def next_iteration(self):
        pass


class AlgorithmBase:

    @staticmethod
    def configure_argparser(argparser):
        pass

    def __init__(self, context, args):
        pass

    async def exit(self, reason):
        pass

    async def handle_exception(self, e):
        raise e

    async def next_iteration(self):
        pass



import ccxt
import ccxt.async_support
import pandas
import os
import json
import functools
import logging
from collections import defaultdict
from enum import auto, Enum
from btrccts.exchange import BacktestExchangeBase
from btrccts.async_exchange import AsyncBacktestExchangeBase
from btrccts.exchange_backend import ExchangeBackend
try:
    import ccxtpro
except ImportError:
    ccxtpro = None
    pass


class StopException(SystemExit):
    pass


class ContextState(Enum):

    BACKTEST = auto()
    LIVE = auto()


class BacktestContext:

    def __init__(self, timeframe, exchange_backends={}):
        self._exchange_backends = defaultdict(functools.partial(
            ExchangeBackend, timeframe=timeframe))
        for key in exchange_backends:
            self._exchange_backends[key] = exchange_backends[key]
        self._timeframe = timeframe

    def create_exchange(self, exchange_id, config={}, async_ccxt=False):
        use_ccxt = ccxt
        base = BacktestExchangeBase
        if async_ccxt:
            use_ccxt = ccxt.async_support
            base = AsyncBacktestExchangeBase
        if exchange_id not in use_ccxt.exchanges:
            raise ValueError('Unknown exchange: {}'.format(exchange_id))
        exchange = getattr(use_ccxt, exchange_id)

        class BacktestExchange(base, exchange):
            pass

        backend = self._exchange_backends[exchange_id]
        instance = BacktestExchange(config=config, exchange_backend=backend)
        return instance

    def date(self):
        return self._timeframe.date()

    def real_date(self):
        return self._timeframe.date()

    def state(self):
        return ContextState.BACKTEST

    def stop(self, msg):
        raise StopException(msg)


class LiveContext:

    def __init__(self, timeframe, conf_dir, auth_aliases={}):
        self._timeframe = timeframe
        self._auth_aliases = auth_aliases
        self._conf_dir = conf_dir

    def create_exchange(self, exchange_id, config={}, async_ccxt=False):
        use_ccxt = ccxt
        if async_ccxt:
            if ccxtpro is not None:
                use_ccxt = ccxtpro
            else:
                use_ccxt = ccxt.async_support
        if exchange_id not in use_ccxt.exchanges:
            raise ValueError('Unknown exchange: {}'.format(exchange_id))
        exchange = getattr(use_ccxt, exchange_id)
        config_file = os.path.join(self._conf_dir, '{}.json'.format(
            self._auth_aliases.get(exchange_id, exchange_id)))
        exchange_config = {'enableRateLimit': True}
        if os.path.isfile(config_file):
            with open(config_file) as f:
                exchange_config.update(json.load(f))
        else:
            logger = logging.getLogger(__package__)
            logger.warning('Config file for exchange {} does not exist: {}'
                           .format(exchange_id, config_file))
        exchange_config.update(config)
        return exchange(exchange_config)

    def date(self):
        return self._timeframe.date()

    def real_date(self):
        return pandas.Timestamp.now(tz='UTC')

    def state(self):
        return ContextState.LIVE

    def stop(self, msg):
        raise StopException(msg)



from ccxt.base.errors import BadRequest
from decimal import Decimal, InvalidOperation


def _convert_float_or_raise(f, msg):
    try:
        val = _convert_float(f)
    except InvalidOperation:
        raise BadRequest('{} needs to be a number'.format(msg))
    if not val.is_finite():
        raise BadRequest('{} needs to be finite'.format(msg))
    return val


def _convert_float(f):
    return Decimal(str(f))



import appdirs
import argparse
import asyncio
import json
import logging
import numpy
import os
import pandas
from asyncio import events, tasks, coroutines
from ccxt.base.errors import NotSupported
from ccxt.base.exchange import Exchange
from enum import Enum, auto
from functools import partial
from btrccts.context import BacktestContext, LiveContext, StopException
from btrccts.exchange_backend import ExchangeBackend
from btrccts.timeframe import Timeframe

USER_CONFIG_DIR = appdirs.user_config_dir(__package__)
USER_DATA_DIR = appdirs.user_data_dir(__package__)
SLEEP_SECONDS = 1
HELP_EPILOG = """
Default config directory: {data}
Default data directory: {config}
""".format(config=USER_CONFIG_DIR, data=USER_DATA_DIR)


async def _run_a_or_sync(func, *args, **kwargs):
    if asyncio.iscoroutinefunction(func):
        return await func(*args, **kwargs)
    else:
        return func(*args, **kwargs)


def _cancel_all_tasks(loop):
    if hasattr(asyncio, 'all_tasks'):
        all_tasks = asyncio.all_tasks
    else:
        all_tasks = asyncio.Task.all_tasks
    to_cancel = all_tasks(loop)
    if not to_cancel:
        return

    for task in to_cancel:
        task.cancel()

    loop.run_until_complete(
        tasks.gather(*to_cancel, loop=loop, return_exceptions=True))

    for task in to_cancel:
        if task.cancelled():
            continue
        if task.exception() is not None:
            loop.call_exception_handler({
                'message': 'unhandled exception during asyncio.run() shutdown',
                'exception': task.exception(),
                'task': task,
            })


def _run_async(main, *, debug=False):
    if events._get_running_loop() is not None:
        raise RuntimeError(
            "asyncio.run() cannot be called from a running event loop")

    if not coroutines.iscoroutine(main):
        raise ValueError("a coroutine was expected, got {!r}".format(main))

    loop = events.new_event_loop()
    try:
        events.set_event_loop(loop)
        loop.set_debug(debug)
        return loop.run_until_complete(main)
    finally:
        try:
            _cancel_all_tasks(loop)
            loop.run_until_complete(loop.shutdown_asyncgens())
        finally:
            events.set_event_loop(None)
            loop.close()


def load_ohlcvs(ohlcv_dir, exchange_names, symbols):
    result = {}
    complete_exchange = False
    if len(symbols) == 0:
        complete_exchange = True
    for exchange_name in exchange_names:
        exchange_result = {}
        result[exchange_name] = exchange_result
        exchange_path = os.path.join(ohlcv_dir, exchange_name)
        if complete_exchange:
            try:
                symbols = []
                for base in os.listdir(exchange_path):
                    quote_path = os.path.join(exchange_path, base)
                    if os.path.isdir(quote_path):
                        for quote in os.listdir(quote_path):
                            if quote.endswith('.csv'):
                                symbols.append(os.path.join(base, quote[:-4]))
            except (FileNotFoundError, NotADirectoryError):
                raise FileNotFoundError(
                    'Cannot find ohlcv directory for exchange ({})'
                    .format(exchange_name))
        for symbol in symbols:
            file_path = os.path.join(exchange_path,
                                     '{}.csv'.format(symbol))
            try:
                ohlcv = pandas.read_csv(
                    file_path,
                    index_col=0, parse_dates=[0], dtype=numpy.float,
                    date_parser=partial(pandas.to_datetime, utc=True)
                    )
                exchange_result[symbol] = ohlcv
            except FileNotFoundError:
                raise FileNotFoundError(
                    'Cannot find symbol ({}) file for exchange ({})'
                    .format(symbol, exchange_name))
            except ValueError:
                raise ValueError(
                    'Cannot parse symbol ({}) file for exchange ({})'
                    .format(symbol, exchange_name))
    return result


class ExitReason(Enum):

    STOPPED = auto()
    EXCEPTION = auto()
    FINISHED = auto()


async def sleep_until(date):
    # Use a loop, so if the system clock changes, we dont sleep too long/short
    while True:
        now = pandas.Timestamp.now(tz='UTC')
        if date <= now:
            return
        sleep_sec = SLEEP_SECONDS
        diff = (date - now).value / 10**9
        if diff < SLEEP_SECONDS:
            sleep_sec = diff
        await asyncio.sleep(sleep_sec)


async def main_loop(timeframe, algorithm, live=False):
    logger = logging.getLogger(__package__)
    logger.info('Starting main_loop')
    while not timeframe.finished():
        try:
            try:
                await _run_a_or_sync(algorithm.next_iteration)
            except (SystemExit, KeyboardInterrupt, StopException,
                    asyncio.CancelledError) as e:
                logger.info('Stopped because of {}: {}'.format(
                    type(e).__name__, e))
                await _run_a_or_sync(algorithm.exit,
                                     reason=ExitReason.STOPPED)
                return algorithm
            except BaseException as e:
                logger.error('Error occured during next_iteration')
                logger.exception(e)
                try:
                    await _run_a_or_sync(algorithm.handle_exception, e)
                except BaseException as e:
                    logger.error(
                        'Exiting because of exception in handle_exception')
                    logger.exception(e)
                    await _run_a_or_sync(algorithm.exit,
                                         reason=ExitReason.EXCEPTION)
                    raise e
            timeframe.add_timedelta()
            if live:
                # We already added a timedelta.
                # If the algo took longer then timedelta,
                # we want to do the next round immediately
                timeframe.add_timedelta_until(pandas.Timestamp.now(tz='UTC'))
                next_date = timeframe.date()
                if next_date is not None:
                    await sleep_until(next_date)
        except (SystemExit, KeyboardInterrupt, asyncio.CancelledError) as e:
            logger.info('Stopped because of {}: {}'.format(
                type(e).__name__, e))
            await _run_a_or_sync(algorithm.exit, reason=ExitReason.STOPPED)
            return algorithm
    await _run_a_or_sync(algorithm.exit, reason=ExitReason.FINISHED)
    logger.info('Finished main_loop')
    return algorithm


def execute_algorithm(exchange_names, symbols, AlgorithmClass, args,
                      start_balances,
                      pd_start_date, pd_end_date, pd_interval,
                      live, auth_aliases,
                      data_dir=USER_DATA_DIR,
                      conf_dir=USER_CONFIG_DIR):
    timeframe = Timeframe(pd_start_date=pd_start_date,
                          pd_end_date=pd_end_date,
                          pd_interval=pd_interval)
    ohlcv_dir = os.path.join(data_dir, 'ohlcv')
    if live:
        context = LiveContext(timeframe=timeframe,
                              conf_dir=conf_dir,
                              auth_aliases=auth_aliases)
    else:
        ohlcvs = load_ohlcvs(ohlcv_dir=ohlcv_dir,
                             exchange_names=exchange_names,
                             symbols=symbols)
        exchange_backends = {}
        for exchange_name in exchange_names:
            exchange_backends[exchange_name] = ExchangeBackend(
                timeframe=timeframe,
                balances=start_balances.get(exchange_name, {}),
                ohlcvs=ohlcvs.get(exchange_name, {}))
        context = BacktestContext(timeframe=timeframe,
                                  exchange_backends=exchange_backends)

    async def func():
        algorithm = AlgorithmClass(context=context,
                                   args=args)
        return await main_loop(timeframe=timeframe,
                               algorithm=algorithm,
                               live=live)
    return _run_async(func())


def parse_params_and_execute_algorithm(AlgorithmClass):
    parser = argparse.ArgumentParser(
        epilog=HELP_EPILOG, formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('--start-date', default='',
                        help='Date to start backtesting, ignored in live mode')
    parser.add_argument('--end-date', default='2009-01-01',
                        help='Date to end backtesting')
    parser.add_argument('--interval', default='1m',
                        help='Timedelta between each iteration')
    parser.add_argument('--exchanges', default='',
                        help='Exchange ids comma separated to load ohlcv')
    parser.add_argument('--symbols', default='',
                        help='Symbols (comma separated) to load ohlcv '
                             'per exchange')
    parser.add_argument('--data-directory', default=USER_DATA_DIR,
                        help='directory where data is stored'
                             ' (e.g. ohlcv data')
    parser.add_argument('--config-directory', default=USER_CONFIG_DIR,
                        help='directory where config is stored'
                             ' (e.g. exchange parameters')
    parser.add_argument('--auth-aliases', default='{}',
                        help='Auth aliases for different exchange'
                             ' config files')
    parser.add_argument('--live', action='store_true',
                        help='Trade live on exchanges')
    parser.add_argument('--start-balances', default='{}',
                        help='Balance at start (json): '
                             '{"exchange": {"BTC": 3}}')
    AlgorithmClass.configure_argparser(parser)
    args = parser.parse_args()
    logger = logging.getLogger(__package__)

    def split_parameters(p):
        if p == '':
            return []
        return p.split(',')

    exchange_names = split_parameters(args.exchanges)
    symbols = split_parameters(args.symbols)
    if not args.live:
        if len(exchange_names) == 0:
            logger.warning('No exchanges specified, do not load ohlcv')
        if len(symbols) == 0:
            logger.warning('No symbols specified, load all ohlcvs per each '
                           'exchange. This can lead to long start times')
    try:
        pd_interval = pandas.Timedelta(
            Exchange.parse_timeframe(args.interval), unit='s')
    except (NotSupported, ValueError):
        raise ValueError('Interval is not valid')
    auth_aliases = {}
    if args.live:
        if args.start_date != '':
            raise ValueError('Start date cannot be set in live mode')
        if args.start_balances != '{}':
            raise ValueError('Start balance cannot be set in live mode')
        pd_start_date = pandas.Timestamp.now(tz='UTC').floor(pd_interval)
        start_balances = None
        auth_aliases = json.loads(args.auth_aliases)
    else:
        pd_start_date = pandas.to_datetime(args.start_date, utc=True)
        start_balances = json.loads(args.start_balances)
    pd_end_date = pandas.to_datetime(args.end_date, utc=True)
    if pandas.isnull(pd_start_date):
        raise ValueError('Start date is not valid')
    if pandas.isnull(pd_end_date):
        raise ValueError('End date is not valid')

    return execute_algorithm(exchange_names=exchange_names,
                             symbols=symbols,
                             pd_start_date=pd_start_date,
                             pd_end_date=pd_end_date,
                             pd_interval=pd_interval,
                             conf_dir=args.config_directory,
                             data_dir=args.data_directory,
                             AlgorithmClass=AlgorithmClass,
                             args=args,
                             auth_aliases=auth_aliases,
                             live=args.live,
                             start_balances=start_balances)



from ccxt.base.errors import InvalidOrder, BadRequest


class AsyncBacktestExchangeBase:

    def __init__(self, config, exchange_backend):
        super().__init__(config=config)
        self._exchange_backend = exchange_backend

    def _check_has(self, name):
        if not self.has[name]:
            raise NotImplementedError('{}: method not implemented: {}'
                                      .format(self.id, name))

    async def cancel_all_orders(self, *args, **kwargs):
        self._check_has('cancelAllOrders')
        raise NotImplementedError('BacktestExchange does not support method '
                                  'cancel_all_orders')

    async def cancel_order(self, id, symbol=None, params={}):
        self._check_has('cancelOrder')
        return self._exchange_backend.cancel_order(id=id, symbol=symbol)

    async def cancel_orders(self, *args, **kwargs):
        self._check_has('cancelOrders')
        raise NotImplementedError('BacktestExchange does not support method '
                                  'cancel_orders')

    async def create_deposit_address(self, *args, **kwargs):
        self._check_has('createDepositAddress')
        raise NotImplementedError('BacktestExchange does not support method '
                                  'create_deposit_address')

    async def create_order(self, symbol, type, side, amount, price=None,
                           params={}):
        self._check_has('createOrder')
        await super().load_markets()
        if type == 'market':
            if not self.has['createMarketOrder']:
                raise NotImplementedError(
                    '{}: method not implemented: createMarketOrder'
                    .format(self.id))
        elif type == 'limit':
            if not self.has['createLimitOrder']:
                raise NotImplementedError(
                    '{}: method not implemented: createLimitOrder'
                    .format(self.id))
        market = self.markets.get(symbol)
        if market is None:
            raise InvalidOrder('Exchange: market does not exist: {}'.format(
                symbol))
        return self._exchange_backend.create_order(
            market=market, side=side, amount=amount, type=type, price=price)

    async def deposit(self, *args, **kwargs):
        self._check_has('deposit')
        raise NotImplementedError('BacktestExchange does not support method '
                                  'deposit')

    async def edit_order(self, *args, **kwargs):
        self._check_has('editOrder')
        raise NotImplementedError('BacktestExchange does not support method '
                                  'edit_order')

    async def fetch_balance(self, params={}):
        self._check_has('fetchBalance')
        result = self._exchange_backend.fetch_balance()
        return self.parse_balance(result)

    async def fetch_closed_orders(
            self, symbol=None, since=None, limit=None, params={}):
        self._check_has('fetchClosedOrders')
        return self._exchange_backend.fetch_closed_orders(
            symbol=symbol, since=since, limit=limit)

    async def fetch_currencies(self, params={}):
        self._check_has('fetchCurrencies')
        return await super().fetch_currencies(params)

    async def fetch_deposit_address(self, *args, **kwargs):
        self._check_has('fetchDepositAddress')
        raise NotImplementedError('BacktestExchange does not support method '
                                  'fetch_deposit_address')

    async def fetch_deposits(self, *args, **kwargs):
        self._check_has('fetchDeposits')
        raise NotImplementedError('BacktestExchange does not support method '
                                  'fetch_deposits')

    async def fetch_l2_order_book(self, *args, **kwargs):
        self._check_has('fetchL2OrderBook')
        raise NotImplementedError('BacktestExchange does not support method '
                                  'fetch_l2_order_book')

    async def fetch_ledger(self, *args, **kwargs):
        self._check_has('fetchLedger')
        raise NotImplementedError('BacktestExchange does not support method '
                                  'fetch_ledger')

    async def fetch_markets(self, params={}):
        self._check_has('fetchMarkets')
        return await super().fetch_markets(params)

    async def fetch_my_trades(self, *args, **kwargs):
        self._check_has('fetchMyTrades')
        raise NotImplementedError('BacktestExchange does not support method '
                                  'fetch_my_trades')

    async def fetch_ohlcv(self, symbol, timeframe='1m', since=None, limit=None,
                          params={}):
        self._check_has('fetchOHLCV')
        if timeframe not in self.timeframes:
            raise BadRequest('Timeframe {} not supported by exchange'.format(
                timeframe))
        data = self._exchange_backend.fetch_ohlcv_dataframe(
            symbol=symbol, timeframe=timeframe, since=since, limit=limit)
        result = [[int(values.Index.value / 10**6),
                   values.open,
                   values.high,
                   values.low,
                   values.close,
                   values.volume] for values in data.itertuples()]
        return result

    async def fetch_open_orders(
            self, symbol=None, since=None, limit=None, params={}):
        self._check_has('fetchOpenOrders')
        return self._exchange_backend.fetch_open_orders(
            symbol=symbol, since=since, limit=limit)

    async def fetch_order(self, id, symbol=None, params={}):
        self._check_has('fetchOrder')
        return self._exchange_backend.fetch_order(id=id, symbol=symbol)

    async def fetch_order_book(self, *args, **kwargs):
        self._check_has('fetchOrderBook')
        raise NotImplementedError('BacktestExchange does not support method '
                                  'fetch_order_book')

    async def fetch_order_books(self, *args, **kwargs):
        self._check_has('fetchOrderBooks')
        raise NotImplementedError('BacktestExchange does not support method '
                                  'fetch_order_books')

    async def fetch_orders(self, *args, **kwargs):
        self._check_has('fetchOrders')
        raise NotImplementedError('BacktestExchange does not support method '
                                  'fetch_orders')

    async def fetch_status(self, *args, **kwargs):
        self._check_has('fetchStatus')
        raise NotImplementedError('BacktestExchange does not support method '
                                  'fetch_status')

    async def fetch_ticker(self, symbol, params={}):
        self._check_has('fetchTicker')
        return self._exchange_backend.fetch_ticker(symbol=symbol)

    async def fetch_tickers(self, *args, **kwargs):
        self._check_has('fetchTickers')
        raise NotImplementedError('BacktestExchange does not support method '
                                  'fetch_tickers')

    async def fetch_time(self, *args, **kwargs):
        self._check_has('fetchTime')
        raise NotImplementedError('BacktestExchange does not support method '
                                  'fetch_time')

    async def fetch_trades(self, *args, **kwargs):
        self._check_has('fetchTrades')
        raise NotImplementedError('BacktestExchange does not support method '
                                  'fetch_trades')

    async def fetch_trading_fee(self, *args, **kwargs):
        self._check_has('fetchTradingFee')
        raise NotImplementedError('BacktestExchange does not support method '
                                  'fetch_trading_fee')

    async def fetch_trading_fees(self, *args, **kwargs):
        self._check_has('fetchTradingFees')
        raise NotImplementedError('BacktestExchange does not support method '
                                  'fetch_trading_fees')

    async def fetch_funding_fee(self, *args, **kwargs):
        self._check_has('fetchFundingFee')
        raise NotImplementedError('BacktestExchange does not support method '
                                  'fetch_funding_fee')

    async def fetch_funding_fees(self, *args, **kwargs):
        self._check_has('fetchFundingFees')
        raise NotImplementedError('BacktestExchange does not support method '
                                  'fetch_funding_fees')

    async def fetch_trading_limits(self, *args, **kwargs):
        self._check_has('fetchTradingLimits')
        raise NotImplementedError('BacktestExchange does not support method '
                                  'fetch_trading_limits')

    async def fetch_transactions(self, *args, **kwargs):
        self._check_has('fetchTransactions')
        raise NotImplementedError('BacktestExchange does not support method '
                                  'fetch_transactions')

    async def fetch_withdrawals(self, *args, **kwargs):
        self._check_has('fetchWithdrawals')
        raise NotImplementedError('BacktestExchange does not support method '
                                  'fetch_withdrawals')

    async def withdraw(self, *args, **kwargs):
        self._check_has('withdraw')
        raise NotImplementedError('BacktestExchange does not support method '
                                  'withdraw')



class Timeframe:

    def __init__(self, pd_start_date, pd_end_date, pd_interval):
        if pd_end_date < pd_start_date:
            raise ValueError('Timeframe: end date is smaller then start date')
        if pd_interval.value <= 0:
            raise ValueError('Timeframe: timedelta needs to be positive')
        self._pd_interval = pd_interval
        self._pd_start_date = pd_start_date
        self._pd_current_date = pd_start_date
        self._pd_end_date = pd_end_date

    def add_timedelta(self):
        self._pd_current_date += self._pd_interval

    def date(self):
        if self.finished():
            return self._pd_end_date
        return self._pd_current_date

    def add_timedelta_until(self, date):
        while self._pd_current_date + self._pd_interval < date:
            self.add_timedelta()

    def start_date(self):
        return self._pd_start_date

    def end_date(self):
        return self._pd_end_date

    def finished(self):
        return self._pd_current_date > self._pd_end_date



import pandas
from ccxt.base.exchange import Exchange
from ccxt.base.errors import BadRequest, InvalidOrder, OrderNotFound
from collections import defaultdict
from copy import deepcopy
from decimal import Decimal
from btrccts.check_dataframe import _check_dataframe
from btrccts.convert_float import _convert_float_or_raise, _convert_float
from btrccts.balance import Balance

DECIMAL_ONE = Decimal('1')


class ExchangeAccount:

    def __init__(self, timeframe, balances={}, ohlcvs={}):
        self._timeframe = timeframe
        self._start_balances = defaultdict(Balance)
        for key in balances:
            self._start_balances[key] = Balance(balances[key])
        self._balances = self._start_balances.copy()
        self._ohlcvs = {}
        for key in ohlcvs:
            self._ohlcvs[key] = _check_dataframe(ohlcvs[key], timeframe)
        self._last_order_id = 0
        self._open_orders = {}
        self._closed_orders = {}
        self._private_order_info = {}
        self._next_private_order_to_update = None

    def _move_to_closed_orders(self, id):
        self._closed_orders[id] = self._open_orders[id]
        del self._open_orders[id]
        del self._private_order_info[id]

    def _update_next_private_order_to_update(self):
        try:
            self._next_private_order_to_update = min(
                filter(lambda x: x['fillable_date'] is not None,
                       self._private_order_info.values()),
                key=lambda x: x['fillable_date'])
        except ValueError:
            self._next_private_order_to_update = None

    def _update_orders(self):
        curr_date = self._timeframe.date()
        while True:
            private_order = self._next_private_order_to_update
            if private_order is None:
                return
            fillable_date = private_order['fillable_date']
            if fillable_date > curr_date:
                return
            order_id = private_order['id']
            timestamp = int(fillable_date.value / 10e5)
            order = self._open_orders[order_id]

            amount = order['amount']
            price = private_order['price']
            base = private_order['base']
            quote = private_order['quote']
            buy = private_order['buy']
            fee_percentage = private_order['fee_percentage']

            self._remove_used_balance(price, amount, base, quote, buy)
            self._update_balance(price, amount, base, quote, buy,
                                 fee_percentage)
            self._fill_order(order, buy, price, timestamp, fee_percentage)
            self._move_to_closed_orders(order_id)

            self._update_next_private_order_to_update()

    def _return_decimal_to_float(self, result):
        for key in result.keys():
            value_type = type(result[key])
            if value_type == Decimal:
                result[key] = float(str(result[key]))
            elif value_type == dict:
                result[key] = self._return_decimal_to_float(result[key])
        return result

    def cancel_order(self, id, symbol=None):
        self._update_orders()
        closed_order = self._closed_orders.get(id)
        if closed_order is not None:
            raise BadRequest('ExchangeAccount: cannot cancel {} order {}'
                             .format(closed_order['status'], id))
        open_order = self._open_orders.get(id)
        if open_order is None:
            raise OrderNotFound('ExchangeAccount: order {} does not exist'
                                .format(id))
        else:
            open_order.update({
                'status': 'canceled',
            })
            private = self._private_order_info[id]
            self._remove_used_balance(amount=open_order['amount'],
                                      price=private['price'],
                                      base=private['base'],
                                      quote=private['quote'],
                                      buy=private['buy'])
            self._move_to_closed_orders(id)
            if private == self._next_private_order_to_update:
                self._update_next_private_order_to_update()
            return {'id': id,
                    'info': {}}

    def create_order(self, market, type, price, side, amount):
        self._update_orders()
        type_market = False
        type_limit = False
        if type == 'market':
            if price is not None:
                raise InvalidOrder(
                    'ExchangeAccount: market order has no price')
            type_market = True
        elif type == 'limit':
            price = _convert_float_or_raise(price, 'ExchangeAccount: price')
            type_limit = True
            if price <= 0:
                raise BadRequest('ExchangeAccount: price needs to be positive')
        else:
            raise InvalidOrder(
                'ExchangeAccount: only market and limit order supported')
        if market is None:
            raise InvalidOrder('ExchangeAccount: market is None')
        symbol = market.get('symbol')
        ohlcv = self._ohlcvs.get(symbol)
        if ohlcv is None:
            raise InvalidOrder('ExchangeAccount: no prices available for {}'
                               .format(symbol))
        if side not in ['buy', 'sell']:
            raise InvalidOrder('ExchangeAccount: side {} not supported'
                               .format(side))
        buy = side == 'buy'
        amount = _convert_float_or_raise(amount, 'ExchangeAccount: amount')
        if amount <= 0:
            raise BadRequest('ExchangeAccount: amount needs to be positive')
        base = market.get('base')
        quote = market.get('quote')
        if base is None:
            raise BadRequest('ExchangeAccount: market has no base')
        if quote is None:
            raise BadRequest('ExchangeAccount: market has no quote')

        self._last_order_id += 1
        order_id = str(self._last_order_id)
        date = self._timeframe.date()
        timestamp = int(date.value / 10e5)
        order = {
            'info': {},
            'id': order_id,
            'timestamp': timestamp,
            'datetime': Exchange.iso8601(timestamp),
            'lastTradeTimestamp': None,
            'symbol': symbol,
            'type': type,
            'side': side,
            'price': None,
            'amount': amount,
            'cost': None,
            'average': None,
            'filled': 0,
            'remaining': amount,
            'status': 'open',
            'fee': {'currency': base if buy else quote,
                    'cost': None,
                    'rate': None},
            'trades': None,
        }

        if type_market:
            # Determinie the price of the market order
            # We could use the next low/high to fill the order, but then we
            # need to wait for the next date to fill the order, otherwise we
            # would introduce a possibility to see the future price
            # (Look-Ahead Bias)
            # If we wait for the next date, we would return a market order that
            # is pending, but this should never happen in reality
            # Maybe the factor should depend on the volume
            factor = Decimal('0.0015')
            if buy:
                price = (1 + factor) * _convert_float(ohlcv['high'][date])
            else:
                price = (1 - factor) * _convert_float(ohlcv['low'][date])
            fee_percentage = market.get('taker', 0)
            fee_percentage = _convert_float_or_raise(fee_percentage,
                                                     'ExchangeAccount: fee')
            self._update_balance(price, amount, base, quote, buy,
                                 fee_percentage)
            self._fill_order(order, buy, price, timestamp, fee_percentage)
            self._closed_orders[order_id] = order
        if type_limit:
            # TODO Probably use taker fee, if the order can be filled now
            fee_percentage = market.get('maker', 0)
            fee_percentage = _convert_float_or_raise(fee_percentage,
                                                     'ExchangeAccount: fee')
            if buy:
                self._balances[quote].change_used(price * amount)
            else:
                self._balances[base].change_used(amount)
            self._open_orders[order_id] = order
            self._private_order_info[order_id] = {
                'id': order_id,
                'base': base,
                'quote': quote,
                'price': price,
                'buy': buy,
                'fee_percentage': fee_percentage,
                'fillable_date': self._limit_order_fillable_date(
                    symbol, buy, price),
            }
            self._update_next_private_order_to_update()

        return {'id': order_id,
                'info': {}}

    def _limit_order_fillable_date(self, symbol, buy, price):
        ohlcv = self._ohlcvs[symbol]
        date = self._timeframe.date()
        if ohlcv.index[0] != date:
            ohlcv = ohlcv[date:]
            # save reduced dataframe for better performance
            self._ohlcvs[symbol] = ohlcv
        # only look at the future
        ohlcv = ohlcv[date + pandas.Timedelta(1, unit='ns'):]
        if buy:
            low = ohlcv.low
            use = low[low <= price]
        else:
            high = ohlcv.high
            use = high[high >= price]
        if use is not None and len(use.index) > 0:
            return use.index[0]
        else:
            return None

    def _update_balance(self, price, amount, base, quote, buy, fee_percentage):
        # First decrease balance, then increase, so
        # decrease can throw and increase wont be affected
        multiplier = DECIMAL_ONE - fee_percentage
        if buy:
            self._balances[quote].change_total(- price * amount)
            self._balances[base].change_total(amount * multiplier)
        else:
            self._balances[base].change_total(- amount)
            self._balances[quote].change_total(price * amount * multiplier)

    def _remove_used_balance(self, price, amount, base, quote, buy):
        if buy:
            self._balances[quote].change_used(- price * amount)
        else:
            self._balances[base].change_used(- amount)

    def _fill_order(self, order, buy, price, timestamp, fee_percentage):
        amount = order['amount']
        amount_price = amount * price
        order.update({
            'average': price,
            'cost': amount_price,
            'filled': amount,
            'lastTradeTimestamp': timestamp,
            'price': price,
            'remaining': 0,
            'status': 'closed',
        })
        order['fee'].update({
            'rate': fee_percentage,
            'cost': fee_percentage * (amount if buy else amount_price),
        })

    def fetch_balance(self):
        self._update_orders()
        result = {}
        for key, balance in self._balances.items():
            result[key] = self._return_decimal_to_float(balance.to_dict())
        return result

    def fetch_order(self, id, symbol=None):
        self._update_orders()
        order = self._closed_orders.get(id)
        if order is None:
            order = self._open_orders.get(id)
        if order is None:
            raise OrderNotFound('ExchangeAccount: order {} does not exist'
                                .format(id))
        return self._return_decimal_to_float(deepcopy(order))

    def _filter_sort_orders(
            self, orders, since, limit, symbol, since_get, filter_non_zero):
        usable_orders = [order for _, order in orders.items()
                         if ((symbol is None or order['symbol'] == symbol) and
                             (filter_non_zero is None or
                              order[filter_non_zero] != 0) and
                             (since is None or order[since_get] > since))]
        usable_orders = sorted(usable_orders, key=lambda x: x[since_get])
        return usable_orders[:limit]

    def fetch_closed_orders(self, symbol=None, since=None, limit=None):
        self._update_orders()
        orders = self._filter_sort_orders(orders=self._closed_orders,
                                          symbol=symbol, limit=limit,
                                          since=since,
                                          filter_non_zero='filled',
                                          since_get='lastTradeTimestamp')
        return [self._return_decimal_to_float(deepcopy(o)) for o in orders]

    def fetch_open_orders(self, symbol=None, since=None, limit=None):
        self._update_orders()
        orders = self._filter_sort_orders(orders=self._open_orders,
                                          symbol=symbol, limit=limit,
                                          since=since,
                                          filter_non_zero=None,
                                          since_get='timestamp')
        return [self._return_decimal_to_float(deepcopy(o)) for o in orders]



from setuptools import setup, find_packages

with open('README.md', encoding='utf-8') as f:
    readme = f.read()


setup(
    name='btrccts',
    version='0.1.0',
    description='BackTest and Run CryptoCurrency Trading Strategies',
    long_description=readme,
    long_description_content_type='text/markdown',
    classifiers=[
        'Programming Language :: Python',
        'Intended Audience :: Developers',
        'Intended Audience :: Financial and Insurance Industry',
        'Topic :: Software Development :: Build Tools',
        'Topic :: Office/Business :: Financial :: Investment',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Operating System :: OS Independent',
        'Environment :: Console'
    ],
    author='Simon Brand',
    author_email='simon.brand@postadigitale.de',
    url='https://github.com/btrccts/btrccts/',
    keywords='btrccts',
    package_dir={'': 'src'},
    packages=find_packages('src/'),
    include_package_data=True,
    zip_safe=False,
    extras_require={
        'dev': ['pycodestyle', 'pyflakes'],
        'ccxt-websockets': ['ccxtpro'],
    },
    install_requires=['ccxt', 'pandas', 'numpy', 'appdirs'],
    entry_points={
        'console_scripts': [
            'btrccts=btrccts:_main',
        ]
    },
)
